#!/bin/env -S deno run -A --unstable-kv

import { assertExists } from "https://deno.land/std@0.223.0/assert/assert_exists.ts";
import { assert } from "https://deno.land/std@0.223.0/assert/assert.ts";

import {
	AutocompleteInteraction,
	ChatInputCommandInteraction,
	Client,
	EmbedBuilder,
	SlashCommandBuilder,
	TextChannel,
} from "npm:discord.js";

import { Channel, channel, collect, days_to_ms, log_from, wait } from "./lib/utils.ts";
import { Storage } from "./lib/storage.ts";
import { Devoir } from "./lib/storage.ts";

const log = log_from(import.meta);

const subjects = new Set([
	"Conception",
	"Hardware",
	"Infrastructure Cloud",
	"Intelligence Artificielle",
	"Robotique",
	"Sécurité",
	"SEPT",
	"Sûreté",
]);

async function main() {
	const debug_server = Deno.env.get("DEBUG");
	if (debug_server !== undefined) log("Starting in debug mode on server", debug_server);
	const [token_file] = Deno.args;
	const token = (await Deno.readTextFile(token_file)).trim();

	const storage = await Storage.open("./local/db");
	const bot = new Client({ intents: ["GuildMessages", "Guilds"] });
	const update_display = channel();

	bot.on("interactionCreate", async (interaction) => {
		if (interaction.isChatInputCommand()) return await handle_command(interaction, storage, update_display);
		if (interaction.isAutocomplete()) return await handle_autocomplete(interaction, storage);
	});

	log("Logging in ...");
	await bot.login(token);
	log("Logged in as", bot.user?.username);

	const commands = build_api_commands(subjects);
	if (debug_server !== undefined) {
		for (const command of commands) command.setName("dbg-" + command.name);
		await bot.application?.commands.set(commands, debug_server);
	} else {
		await bot.application?.commands.set(commands);
	}
	log("Registered", commands.length, "commands.");

	update_loop(bot, storage, update_display);
	update_display.send();

	notification_loop(bot, storage);
	log("Oki.");
}

async function fetch_feed_channel(bot: Client<boolean>, feed: { channel_id: string }) {
	const feed_channel = await bot.channels.fetch(feed.channel_id);
	assertExists(feed_channel);
	assert(feed_channel instanceof TextChannel);
	return feed_channel;
}

function build_api_commands(subjects: Set<string>) {
	const subjects_as_choices = [...subjects.values()].map((name) => ({ name, value: name }));
	const devoir_command = new SlashCommandBuilder()
		.setName("devoir")
		.setDescription("Manipuler les devoirs à effectuer.")
		.addSubcommand((command) =>
			command.setName("ajouter")
				.setDescription("Ajouter un nouveau devoir à effectuer.")
				.addStringOption((option) =>
					option.setName("matière")
						.setDescription("Matière du devoir à effectuer.")
						.addChoices(...subjects_as_choices)
						.setRequired(true)
				)
				.addNumberOption((option) =>
					option.setName("jours")
						.setMinValue(0)
						.setDescription("Nombre de jours avant la date du rendu.")
						.setRequired(true)
				)
				.addStringOption((option) =>
					option.setName("description")
						.setDescription("Description du devoir à effectuer.")
						.setRequired(true)
				)
		).addSubcommand((command) =>
			command.setName("retirer")
				.setDescription("Retirer un devoir.")
				.addStringOption((option) =>
					option.setName("devoir")
						.setDescription("Devoir à retirer.")
						.setAutocomplete(true)
						.setRequired(true)
				)
		).addSubcommand((command) =>
			command.setName("éditer")
				.setDescription("Éditer un devoir.")
				.addStringOption((option) =>
					option.setName("devoir")
						.setDescription("Devoir à éditer.")
						.setAutocomplete(true)
						.setRequired(true)
				)
				.addStringOption((option) =>
					option.setName("matière")
						.setDescription("Nouvelle matière.")
						.addChoices(...subjects_as_choices)
						.setRequired(false)
				)
				.addNumberOption((option) =>
					option.setName("jours")
						.setMinValue(0)
						.setDescription("Nouveau nombre de jours avant la date du rendu.")
						.setRequired(false)
				)
				.addStringOption((option) =>
					option.setName("description")
						.setDescription("Nouvelle description.")
						.setRequired(false)
				)
		);
	const adm_command = new SlashCommandBuilder().setName("adm")
		.setDescription("Commandes d'administration.")
		.addSubcommand((command) =>
			command.setName("ajouter-feed")
				.setDescription("Ajoute un salon comme feed de notifications.")
				.addChannelOption((option) =>
					option.setName("salon")
						.setDescription("Salon du nouveau feed.")
						.setRequired(true)
				)
		);
	return [devoir_command, adm_command];
}

async function handle_command(
	interaction: ChatInputCommandInteraction,
	storage: Storage,
	update_display: Channel,
) {
	log("Received command", interaction.commandName);
	if (interaction.commandName === "devoir") {
		const subcommand = interaction.options.getSubcommand(true);
		if (subcommand === "ajouter") {
			const subject = interaction.options.getString("matière", true);
			const days = interaction.options.getNumber("jours", true);
			const description = interaction.options.getString("description", true);
			const date = Date.now() + days_to_ms(days);
			const { id } = await storage.devoirs.add({ date, description, subject });
			const embed = new EmbedBuilder()
				.setTitle("`🌟` Ajouté")
				.setDescription(`Nouveau devoir ajouté avec succès.`)
				.setFooter({ text: `(devoir id:${id})` });
			await interaction.reply({ embeds: [embed], ephemeral: true });
			update_display.send();
			return;
		}
		if (subcommand === "retirer") {
			const id = interaction.options.getString("devoir", true);
			await storage.devoirs.delete({ id });
			const embed = new EmbedBuilder()
				.setTitle("`🗑️` Retiré")
				.setDescription(`Devoir supprimé avec succès.`)
				.setFooter({ text: `(devoir id:${id})` });
			await interaction.reply({ embeds: [embed], ephemeral: true });
			update_display.send();
			return;
		}
		if (subcommand === "éditer") {
			const id = interaction.options.getString("devoir", true);
			const subject = interaction.options.getString("matière", false);
			if (subject !== null) storage.devoirs.update({ id }, (d) => d.subject = subject);
			const description = interaction.options.getString("description", false);
			if (description !== null) storage.devoirs.update({ id }, (d) => d.description = description);
			const days = interaction.options.getNumber("jours", false);
			if (days !== null) storage.devoirs.update({ id }, (d) => d.date = Date.now() + days_to_ms(days));
			const embed = new EmbedBuilder()
				.setTitle("`🔧` Édité")
				.setDescription(`Devoir mis à jour avec succès.`)
				.setFooter({ text: `(devoir id:${id})` });
			await interaction.reply({ embeds: [embed], ephemeral: true });
			update_display.send();
			return;
		}
		return log("Unknown devoir sub command", subcommand);
	}
	if (interaction.commandName === "adm") {
		const subcommand = interaction.options.getSubcommand(true);
		if (subcommand === "ajouter-feed") {
			const channel = interaction.options.getChannel("salon", true);
			const is_text_channel = channel instanceof TextChannel;
			if (!is_text_channel) {
				await interaction.reply("Channel must be text.");
				return;
			}
			const board_message = await channel.send("[board]");
			const feed_id = await storage.feeds.add({
				channel_id: channel.id,
				board_message_id: board_message.id,
				notification_ids: new Set(),
			});
			await interaction.reply("Added feed " + feed_id.id);
		}
		return log("Unknown adm sub command", subcommand);
	}
	log("Unknown command", interaction.commandName);
}

async function handle_autocomplete(interaction: AutocompleteInteraction, storage: Storage) {
	log("Auto completing.");
	if (interaction.commandName === "devoir") {
		const subcommand = interaction.options.getSubcommand(true);
		if (subcommand === "retirer") {
			const devoirs = await collect(storage.devoirs.list());
			const mapped = devoirs.map(([{ id }, value]) => ({
				name: trimmed(`[${value.subject}] ${value.description}`, 100),
				value: id,
			}));
			return await interaction.respond(mapped);
		}
		if (subcommand === "éditer") {
			const devoirs = await collect(storage.devoirs.list());
			const mapped = devoirs.map(([{ id }, value]) => ({
				name: trimmed(`[${value.subject}] ${value.description}`, 100),
				value: id,
			}));
			return await interaction.respond(mapped);
		}
		return log("Unknown devoir sub command", subcommand);
	}
	log("Unknown command", interaction.commandName);
}

async function update_loop(bot: Client, storage: Storage, update_display: Channel) {
	log("Waiting for updates.");
	while (true) {
		const _trigger = await update_display.receive();
		log("Updating board.");
		for await (const [_, feed] of storage.feeds.list()) {
			const feed_channel = await fetch_feed_channel(bot, feed);
			const embed = new EmbedBuilder().setTitle("`📚` Devoirs")
				.setFooter({ text: "Mise à jour " + Date.now() });
			const sorted_devoirs = (await (collect(storage.devoirs.list())))
				.map(([_, d]) => d).toSorted((a, b) => a.date > b.date ? 1 : -1);
			for (const devoir of sorted_devoirs) {
				embed.addFields({ name: format_devoir_title(devoir), value: devoir.description });
			}
			const board_message = await feed_channel.messages.fetch(feed.board_message_id);
			board_message.edit({ embeds: [embed], content: "" });
		}
	}
}

const _1min = 60 * 1000;
const _24h = 24 * 60 * _1min;
const _7d = 7 * _24h;

function format_devoir_title(devoir: Devoir) {
	const date = new Date(devoir.date);
	return `${date.getDate()}/${date.getMonth() + 1} - ${devoir.subject}`;
}

async function notification_loop(bot: Client, storage: Storage) {
	while (true) {
		// get all devoirs to notify
		const devoirs_to_notify = new Set<string>();
		const notification_threshold = _7d;
		for await (const [devoir_id, devoir] of storage.devoirs.list()) {
			const time_left = devoir.date - Date.now();
			if (time_left < 0) continue; // TODO : delete devoir.
			if (time_left > notification_threshold) continue;
			devoirs_to_notify.add(devoir_id.id);
		}

		// delete all obsolete notifications
		for await (const [notification_id, notification] of storage.notifications.list()) {
			if (devoirs_to_notify.has(notification.devoir_id)) continue;
			const feed = await storage.feeds.get({ id: notification.feed_id });
			if (feed === null) continue;
			const feed_channel = await fetch_feed_channel(bot, feed);
			try {
				const message = await feed_channel.messages.fetch(notification.message_id);
				await message.delete();
			} catch (_) { /* . */ }
			await storage.notifications.delete(notification_id);
			await storage.feeds.update({ id: notification.feed_id }, (f) => f.notification_ids.delete(notification_id.id));
		}

		// create missing messages.
		for await (const [feed_id, feed] of storage.feeds.list()) {
			const feed_channel = await fetch_feed_channel(bot, feed);
			// find devoirs needing to create a notification for
			const devoirs_to_notify_in_feed = new Set(devoirs_to_notify.values());
			for (const existing_notification_id of feed.notification_ids.values()) {
				const notification = await storage.notifications.get({ id: existing_notification_id });
				assertExists(notification);
				devoirs_to_notify_in_feed.delete(notification.devoir_id);
			}
			// create notifications
			for (const devoir_id of devoirs_to_notify_in_feed.values()) {
				const devoir = await storage.devoirs.get({ id: devoir_id });
				if (devoir === null) continue;
				const embed = new EmbedBuilder()
					.setTitle(format_devoir_title(devoir))
					.setDescription(devoir.description);
				const message = await feed_channel.send({ embeds: [embed] });
				const notification_id = await storage.notifications.add({
					devoir_id: devoir_id,
					message_id: message.id,
					feed_id: feed_id.id,
				});
				storage.feeds.update(feed_id, (f) => f.notification_ids.add(notification_id.id));
			}
		}

		// for each notification, update.
		for await (const [_, notification] of storage.notifications.list()) {
			const devoir = await storage.devoirs.get({ id: notification.devoir_id });
			if (devoir === null) continue;
			const feed = await storage.feeds.get({ id: notification.feed_id });
			if (feed === null) continue;
			const feed_channel = await fetch_feed_channel(bot, feed);
			const message = await feed_channel.messages.fetch(notification.message_id);
			const ms_left = devoir.date - Date.now();
			const days_left = Math.floor(ms_left / _24h);
			const embed = new EmbedBuilder()
				.setTitle(`${format_devoir_title(devoir)} (${days_left} jours)`)
				.setDescription(devoir.description)
				.setColor(0x8888ff);
			if (days_left <= 5) embed.setColor(0xffd726);
			if (days_left <= 3) embed.setColor(0xff8b26);
			if (days_left <= 1) embed.setColor(0xff2626);
			await message.edit({ embeds: [embed], content: "" });
		}

		await wait(15 * _1min);
	}
}

function trimmed(text: string, width: number) {
	if (text.length < width) return text;
	else return text.slice(0, width - 4) + "...";
}

if (import.meta.main) await main();
